"""
module catan provides classes and enums useful for representing a catan game.

The main class is Game, which contains Players, a Board, and a CatanLog.

See module boardbuilder for the mechanics of creating and modifying Board objects.

All classes in this module:
- Game
- Player
- Board
- Tile
- Terrain
- HexNumber
- Port
- Piece
- PieceType
"""



from enum import Enum


class PieceType(Enum):
    settlement = 'settlement'
    road = 'road'
    city = 'city'
    robber = 'robber'


class Piece(object):
    """
    class Piece represents a single game piece on the board.

    Allowed types are described in enum PieceType
    """
    def __init__(self, type, owner):
        self.type = type
        self.owner = owner

    def __repr__(self):
        return '<Piece type={}, owner={}>'.format(self.type.value, self.owner)



"""
module boardbuilder is responsible for creating starting board layouts.

It can create a variety of boards by supplying various options.
- Options: [terrain, numbers, ports, pieces, players]
- Option values: [Opt.empty, Opt.random, Opt.preset, Opt.debug]

The default options are defined in #get_opts.

Use #get_opts to convert a dictionary mapping str->str to a dictionary
mapping str->Opts. #get_opts will also apply the default option values
for each option not supplied.

Use #build to build a new board with the passed options.

Use #modify to modify an existing board instead of building a new one.
This will reset the board. #reset is an alias.
"""
from enum import Enum
import logging
import pprint
import random
import hexgrid
import catan.game
import catan.states
import catan.board
import catan.pieces


class Opt(Enum):
    empty = 'empty'
    random = 'random'
    preset = 'preset'
    debug = 'debug'

    def __repr__(self):
        return 'opt:{}'.format(self.value)


def get_opts(opts):
    """
    Validate options and apply defaults for options not supplied.

    :param opts: dictionary mapping str->str.
    :return: dictionary mapping str->Opt. All possible keys are present.
    """
    defaults = {
        'terrain': Opt.empty,
        'numbers': Opt.empty,
        'ports': Opt.preset,
        'pieces': Opt.preset,
        'players': Opt.preset,
    }
    _opts = defaults.copy()
    if opts is None:
        opts = dict()
    try:
        for key, val in opts.copy().items():
            opts[key] = Opt(val)
        _opts.update(opts)
    except Exception:
        raise ValueError('Invalid options={}'.format(opts))
    logging.debug('used defaults=\n{}\n on opts=\n{}\nreturned total opts=\n{}'.format(
        pprint.pformat(defaults),
        pprint.pformat(opts),
        pprint.pformat(_opts)))
    return _opts


def build(opts=None):
    """
    Build a new board using the given options.
    :param opts: dictionary mapping str->Opt
    :return: the new board, Board
    """
    board = catan.board.Board()
    modify(board, opts)
    return board


def reset(board, opts=None):
    """
    Alias for #modify. Resets an existing board.
    """
    modify(board, opts)
    return None


def modify(board, opts=None):
    """
    Reset an existing board using the given options.
    :param board: the board to reset
    :param opts: dictionary mapping str->Opt
    :return: None
    """
    opts = get_opts(opts)
    board.tiles = _generate_tiles(opts['terrain'], opts['numbers'])
    board.ports = _generate_ports(opts['ports'])
    board.state = catan.states.BoardStateModifiable(board)
    board.pieces = _generate_pieces(board.tiles, board.ports, opts['players'], opts['pieces'])
    return None


def _generate_tiles(terrain_opts, numbers_opts):
    """
    Generate a list of tiles using the given terrain and numbers options.

    terrain options supported:
    - Opt.empty -> all tiles are desert
    - Opt.random -> tiles are randomized
    - Opt.preset ->
    - Opt.debug -> alias for Opt.random

    numbers options supported:
    - Opt.empty -> no tiles have numbers
    - Opt.random -> numbers are randomized
    - Opt.preset ->
    - Opt.debug -> alias for Opt.random

    :param terrain_opts: Opt
    :param numbers_opts: Opt
    :return: list(Tile)
    """
    terrain = None
    numbers = None

    if terrain_opts == Opt.empty:
        terrain = ([catan.board.Terrain.desert] * catan.board.NUM_TILES)
    elif terrain_opts in (Opt.random, Opt.debug):
        terrain = ([catan.board.Terrain.desert] +
                   [catan.board.Terrain.brick] * 3 +
                   [catan.board.Terrain.ore] * 3 +
                   [catan.board.Terrain.wood] * 4 +
                   [catan.board.Terrain.sheep] * 4 +
                   [catan.board.Terrain.wheat] * 4)
        random.shuffle(terrain)
    elif terrain_opts == Opt.preset:
        terrain = ([catan.board.Terrain.desert] * catan.board.NUM_TILES)
        logging.warning('Preset terrain option not yet implemented')

    if numbers_opts == Opt.empty:
        numbers = ([catan.board.HexNumber.none] * catan.board.NUM_TILES)
    elif numbers_opts in (Opt.random, Opt.debug):
        numbers = ([catan.board.HexNumber.two] +
                   [catan.board.HexNumber.three]*2 + [catan.board.HexNumber.four]*2 +
                   [catan.board.HexNumber.five]*2 + [catan.board.HexNumber.six]*2 +
                   [catan.board.HexNumber.eight]*2 + [catan.board.HexNumber.nine]*2 +
                   [catan.board.HexNumber.ten]*2 + [catan.board.HexNumber.eleven]*2 +
                   [catan.board.HexNumber.twelve])
        random.shuffle(numbers)
        numbers.insert(terrain.index(catan.board.Terrain.desert), catan.board.HexNumber.none)
    elif numbers_opts == Opt.preset:
        numbers = ([catan.board.HexNumber.none] * catan.board.NUM_TILES)
        logging.warning('Preset numbers option not yet implemented')

    assert len(numbers) == catan.board.NUM_TILES
    assert len(terrain) == catan.board.NUM_TILES

    tile_data = list(zip(terrain, numbers))
    tiles = [catan.board.Tile(i, t, n) for i, (t, n) in enumerate(tile_data, 1)]

    return tiles


def _generate_ports(port_opts):
    """
    Generate a list of ports using the given options.

    port options supported:
    - Opt.empty ->
    - Opt.random ->
    - Opt.preset -> ports are in default locations
    - Opt.debug -> alias for Opt.preset

    :param port_opts: Opt
    :return: list(Port)
    """
    if port_opts in [Opt.preset, Opt.debug]:
        _preset_ports = [(1, 'NW', catan.board.PortType.any3),
                         (2, 'W', catan.board.PortType.wood),
                         (4, 'W', catan.board.PortType.brick),
                         (5, 'SW', catan.board.PortType.any3),
                         (6, 'SE', catan.board.PortType.any3),
                         (8, 'SE', catan.board.PortType.sheep),
                         (9, 'E', catan.board.PortType.any3),
                         (10, 'NE', catan.board.PortType.ore),
                         (12, 'NE', catan.board.PortType.wheat)]
        return [catan.board.Port(tile, dir, port_type)
                for tile, dir, port_type in _preset_ports]
    elif port_opts in [Opt.empty, Opt.random]:
        logging.warning('{} option not yet implemented'.format(port_opts))
        return []


def _generate_pieces(tiles, ports, players_opts, pieces_opts):
    """
    Generate a dictionary of pieces using the given options.

    pieces options supported:
    - Opt.empty -> no locations have pieces
    - Opt.random ->
    - Opt.preset -> robber is placed on the first desert found
    - Opt.debug -> a variety of pieces are placed around the board

    :param tiles: list of tiles from _generate_tiles
    :param ports: list of ports from _generate_ports
    :param players_opts: Opt
    :param pieces_opts: Opt
    :return: dictionary mapping (hexgrid.TYPE, coord:int) -> Piece
    """
    if pieces_opts == Opt.empty:
        return dict()
    elif pieces_opts == Opt.debug:
        players = catan.game.Game.get_debug_players()
        return {
            (hexgrid.NODE, 0x23): catan.pieces.Piece(catan.pieces.PieceType.settlement, players[0]),
            (hexgrid.EDGE, 0x22): catan.pieces.Piece(catan.pieces.PieceType.road, players[0]),
            (hexgrid.NODE, 0x67): catan.pieces.Piece(catan.pieces.PieceType.settlement, players[1]),
            (hexgrid.EDGE, 0x98): catan.pieces.Piece(catan.pieces.PieceType.road, players[1]),
            (hexgrid.NODE, 0x87): catan.pieces.Piece(catan.pieces.PieceType.settlement, players[2]),
            (hexgrid.EDGE, 0x89): catan.pieces.Piece(catan.pieces.PieceType.road, players[2]),
            (hexgrid.EDGE, 0xA9): catan.pieces.Piece(catan.pieces.PieceType.road, players[3]),
            (hexgrid.TILE, 0x77): catan.pieces.Piece(catan.pieces.PieceType.robber, None),
        }
    elif pieces_opts in (Opt.preset, ):
        deserts = filter(lambda tile: tile.terrain == catan.board.Terrain.desert, tiles)
        coord = hexgrid.tile_id_to_coord(list(deserts)[0].tile_id)
        return {
            (hexgrid.TILE, coord): catan.pieces.Piece(catan.pieces.PieceType.robber, None)
        }
    elif pieces_opts in (Opt.random, ):
        logging.warning('{} option not yet implemented'.format(pieces_opts))


def _check_red_placement(tiles):
    """
    Returns True if no red numbers are on adjacent tiles.
    Returns False if any red numbers are on adjacent tiles.

    Not yet implemented.
    """
    logging.warning('"Check red placement" not yet implemented')


import copy
from enum import Enum
import logging
import hexgrid
from catan import boardbuilder, states
from catan.pieces import PieceType, Piece


class Board(object):
    """
    class Board represents a catan board. It has tiles, ports, and pieces.

    A Board has pieces, which is a dictionary mapping (hexgrid.TYPE, coord) -> Piece.

    Use #place_piece, #move_piece, and #remove_piece to manage pieces on the board.

    Use #get_pieces to get all the pieces at a particular coordinate of the allowed types.
    """
    def __init__(self, terrain=None, numbers=None, ports=None, pieces=None, players=None):
        """
        Create a new board. Creation will be delegated to module boardbuilder.

        :param terrain: terrain option, boardbuilder.Opt
        :param numbers: numbers option, boardbuilder.Opt
        :param ports: ports option, boardbuilder.Opt
        :param pieces: pieces option, boardbuilder.Opt
        :param players: players option, boardbuilder.Opt
        """
        self.tiles = list()
        self.ports = list()
        self.state = states.BoardState(self)
        self.pieces = dict()

        self.opts = dict()
        if terrain is not None:
            self.opts['terrain'] = terrain
        if numbers is not None:
            self.opts['numbers'] = numbers
        if ports is not None:
            self.opts['ports'] = ports
        if pieces is not None:
            self.opts['pieces'] = pieces
        if players is not None:
            self.opts['players'] = players

        self.reset()
        self.observers = set()

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = object.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == 'observers':
                setattr(result, k, set(v))
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def restore(self, board):
        """
        Restore this Board object to match the properties and state of the given Board object
        :param board: properties to restore to the current (self) Board
        """
        self.tiles = board.tiles
        self.ports = board.ports

        self.state = board.state
        self.state.board = self

        self.pieces = board.pieces
        self.opts = board.opts
        self.observers = board.observers

        self.notify_observers()

    def notify_observers(self):
        for obs in self.observers:
            obs.notify(self)

    def lock(self):
        self.state = states.BoardStateLocked(self)
        for port in self.ports.copy():
            if port.type == PortType.none:
                self.ports.remove(port)
        self.notify_observers()

    def unlock(self):
        self.state = states.BoardStateModifiable(self)

    def reset(self, terrain=None, numbers=None, ports=None, pieces=None, players=None):
        opts = self.opts.copy()
        if terrain is not None:
            opts['terrain'] = terrain
        if numbers is not None:
            opts['numbers'] = numbers
        if ports is not None:
            opts['ports'] = ports
        if pieces is not None:
            opts['pieces'] = pieces
        if players is not None:
            opts['players'] = players
        boardbuilder.reset(self, opts=opts)

    def can_place_piece(self, piece, coord):
        if piece.type == PieceType.road:
            logging.warning('"Can place road" not yet implemented')
            return True
        elif piece.type == PieceType.settlement:
            logging.warning('"Can place settlement" not yet implemented')
            return True
        elif piece.type == PieceType.city:
            logging.warning('"Can place city" not yet implemented')
            return True
        elif piece.type == PieceType.robber:
            logging.warning('"Can place robber" not yet implemented')
            return True
        else:
            logging.debug('Can\'t place piece={} on coord={}'.format(
                piece.value, hex(coord)
            ))
            return self.pieces.get(coord) is None

    def place_piece(self, piece, coord):
        if not self.can_place_piece(piece, coord):
            logging.critical('ILLEGAL: Attempted to place piece={} on coord={}'.format(
                piece.value, hex(coord)
            ))
        logging.debug('Placed piece={} on coord={}'.format(
            piece, hex(coord)
        ))
        hex_type = self._piece_type_to_hex_type(piece.type)
        self.pieces[(hex_type, coord)] = piece

    def move_piece(self, piece, from_coord, to_coord):
        from_index = (self._piece_type_to_hex_type(piece.type), from_coord)
        if from_index not in self.pieces:
            logging.warning('Attempted to move piece={} which was NOT on the board'.format(from_index))
            return
        self.place_piece(piece, to_coord)
        self.remove_piece(piece, from_coord)

    def remove_piece(self, piece, coord):
        index = (self._piece_type_to_hex_type(piece.type), coord)
        try:
            self.pieces.pop(index)
            logging.debug('Removed piece={}'.format(index))
        except ValueError:
            logging.critical('Attempted to remove piece={} which was NOT on the board'.format(index))

    def get_pieces(self, types=tuple(), coord=None):
        if coord is None:
            logging.critical('Attempted to get_piece with coord={}'.format(coord))
            return Piece(None, None)
        indexes = set((self._piece_type_to_hex_type(t), coord) for t in types)
        pieces = [self.pieces[idx] for idx in indexes if idx in self.pieces]
        if len(pieces) == 0:
            #logging.warning('Found zero pieces at {}'.format(indexes))
            pass
        elif len(pieces) == 1:
            logging.debug('Found one piece at {}: {}'.format(indexes, pieces[0]))
        elif len(pieces) > 1:
            logging.debug('Found {} pieces at {}: {}'.format(len(pieces), indexes, coord, pieces))
        return pieces

    def get_port_at(self, tile_id, direction):
        """
        If no port is found, a new none port is made and added to self.ports.

        Returns the port.

        :param tile_id:
        :param direction:
        :return: Port
        """
        for port in self.ports:
            if port.tile_id == tile_id and port.direction == direction:
                return port
        port = Port(tile_id, direction, PortType.none)
        self.ports.append(port)
        return port

    def _piece_type_to_hex_type(self, piece_type):
        if piece_type in (PieceType.road, ):
            return hexgrid.EDGE
        elif piece_type in (PieceType.settlement, PieceType.city):
            return hexgrid.NODE
        elif piece_type in (PieceType.robber, ):
            return hexgrid.TILE
        else:
            logging.critical('piece type={} has no corresponding hex type. Returning None'.format(piece_type))
            return None

    def cycle_hex_type(self, tile_id):
        if self.state.modifiable():
            tile = self.tiles[tile_id - 1]
            next_idx = (list(Terrain).index(tile.terrain) + 1) % len(Terrain)
            next_terrain = list(Terrain)[next_idx]
            tile.terrain = next_terrain
        else:
            logging.debug('Attempted to cycle terrain on tile={} on a locked board'.format(tile_id))
        self.notify_observers()

    def cycle_hex_number(self, tile_id):
        if self.state.modifiable():
            tile = self.tiles[tile_id - 1]
            next_idx = (list(HexNumber).index(tile.number) + 1) % len(HexNumber)
            next_hex_number = list(HexNumber)[next_idx]
            tile.number = next_hex_number
        else:
            logging.debug('Attempted to cycle number on tile={} on a locked board'.format(tile_id))
        self.notify_observers()

    def cycle_port_type(self, tile_id, direction):
        if self.state.modifiable():
            port = self.get_port_at(tile_id, direction)
            port.type = PortType.next_ui(port.type)
        else:
            logging.debug('Attempted to cycle port on coord=({},{}) on a locked board'.format(tile_id, direction))
        self.notify_observers()

    def rotate_ports(self):
        """
        Rotates the ports 90 degrees. Useful when using the default port setup but the spectator is watching
        at a "rotated" angle from "true north".
        """
        for port in self.ports:
            port.tile_id = ((port.tile_id + 1) % len(hexgrid.coastal_tile_ids())) + 1
            port.direction = hexgrid.rotate_direction(hexgrid.EDGE, port.direction, ccw=True)
        self.notify_observers()

class Tile(object):
    """
    class Tile represents a hex tile on the catan board.

    It contains a tile identifier, a terrain type, and a number.
    """
    def __init__(self, tile_id, terrain, number):
        """
        :param tile_id: tile identifier, int, see module hexgrid
        :param terrain: Terrain
        :param number: HexNumber
        :return:
        """
        self.tile_id = tile_id
        self.terrain = terrain
        self.number = number

# Number of tiles on the catan board. This should probably be in module hexgrid.
NUM_TILES = 3+4+5+4+3


class Terrain(Enum):
    wood = 'wood'
    brick = 'brick'
    wheat = 'wheat'
    sheep = 'sheep'
    ore = 'ore'
    desert = 'desert'

    def __repr__(self):
        return self.value


class HexNumber(Enum):
    none = None
    two = 2
    three = 3
    four = 4
    five = 5
    six = 6
    eight = 8
    nine = 9
    ten = 10
    eleven = 11
    twelve = 12


class PortType(Enum):
    any4 = '4:1' # not used in UI, only used in trading
    any3 = '3:1'
    wood = 'wood'
    brick = 'brick'
    wheat = 'wheat'
    sheep = 'sheep'
    ore = 'ore'
    none = 'none' # only used in UI, not used in trading

    @classmethod
    def list_ui(cls):
        return list(filter(lambda pt: pt != PortType.any4, PortType))

    @classmethod
    def list_trading(cls):
        return list(filter(lambda pt: pt != PortType.none, PortType))

    @classmethod
    def next_ui(cls, ptype):
        types = list(PortType)
        next_idx = (types.index(ptype) + 1) % len(types)
        next_port_type = types[next_idx]
        if next_port_type == PortType.any4:
            next_port_type = PortType.next_ui(next_port_type)
        return next_port_type


class Port(object):
    """
    class Port represents a single port on the board.

    Allowed types are described in enum PortType.
    """
    def __init__(self, tile_id, direction, type):
        self.tile_id = tile_id
        self.direction = direction
        self.type = type

    def __repr__(self):
        return '{}({},{})'.format(self.type.value, self.tile_id, self.direction)




"""
module states provides catan state machines which semi-correctly implement the State Pattern

State Pattern: https://en.wikipedia.org/wiki/State_pattern

The Game has a state whose type is one of the GameState types defined in this module.
The Game has a dev card state whose type is one of the DevCardPlayabilityState types defined in this module.
The Board has a state whose type is one of the BoardState types defined in this module.

Each state machine is described in base state's docstring.

Actions
-------

Callers should invoke action methods on the object directly, and the object will delegate
actions to its state as necessary.

e.g.
    # class Game
    def steal(self, victim):
        if victim is None:
            victim = Player(1, 'nobody', 'nobody')
        self.state.steal(victim)
    # class GameStateSteal
    def steal(self, victim):
        self.game.catanlog.log_player_moves_robber_and_steals(
            self.game.get_cur_player(),
            self.game.robber_tile,
            victim
        )
        self.game.set_state(GameStateDuringTurnAfterRoll(self.game))
    # class GameStateStealUsingKnight
    def steal(self, victim):
        self.game.catanlog.log_player_plays_dev_knight(
            self.game.get_cur_player(),
            self.game.robber_tile,
            victim
        )
        self.game.set_state(GameStateDuringTurnAfterRoll(self.game))

State Capabilities
------------------

Callers should query state capabilities through the state.

e.g.
    if game.state.can_trade():
        tradingUI.show()
    else:
        tradingUI.hide()

Any new state capabilities must be named like can_do_xyz() and must return True or False.
When a GameState subclass doesn't implement can_do_xyz2(), the method call will be caught in
GameState.__getattr__. The method call will be ignored and None will be returned instead.

If the method does not look like can_do_xyz(), it will be logged.

"""
import logging
import hexgrid
import catan.pieces


class GameState(object):
    """
    class GameState is the base game state. All game states inherit from GameState.

    sub-states are always allowed to override provided methods.

    this state implements:
        None
    this state provides:
        None
    sub-states must implement:
        is_in_game()
    """
    def __init__(self, game):
        self.game = game

    def __getattr__(self, name):
        """Return false for methods called on GameStates which don't have those methods.
        This should be ok, since __getattr__ is only called as a last resort
        i.e. if there are no attributes in the instance that match the name

        source: http://stackoverflow.com/a/2405617/1817465
        """
        def method(*args):
            return None
        if 'can_' not in name:
            # can_do_xyz methods are ok to return None if not implemented
            logging.debug('Method {0} not found'.format(name))
        return method

    def is_in_game(self):
        """
        See GameStateInGame for details.

        :return Boolean
        """
        pass


class GameStateNotInGame(GameState):
    """
    All NOT-IN-GAME states inherit from this state.

    See GameStateInGame for details.

    this state implements:
        is_in_game()
    this state provides:
        None
    sub-classes must implement:
        None
    """
    def is_in_game(self):
        return False


class GameStateNotInGameMoveRobber(GameStateNotInGame):
    """
    Moving the robber while setting up the board.
    """
    def can_move_robber(self):
        return True

    def move_robber(self, tile_id):
        robbers = self.game.board.get_pieces((catan.pieces.PieceType.robber, ),
                                             hexgrid.tile_id_to_coord(self.game.robber_tile))
        to_coord = hexgrid.tile_id_to_coord(tile_id)
        if robbers:
            robber = robbers[0]
            from_coord = hexgrid.tile_id_to_coord(self.game.robber_tile)
            self.game.board.move_piece(robber, from_coord, to_coord)
        else:
            robber = catan.pieces.Piece(catan.pieces.PieceType.robber, None)
            self.game.board.place_piece(robber, to_coord)
        if len(robbers) != 1:
            logging.warning('{} robbers found in board.pieces'.format(len(robbers)))
        self.game.robber_tile = tile_id
        self.game.set_state(GameStateNotInGame(self.game))


class GameStateInGame(GameState):
    """
    All IN-GAME states inherit from this state.

    In game is defined as taking turns, rolling dice, placing pieces, etc.
    In game starts on 'Start Game', and ends on 'End Game'

    this state implements:
        is_in_game()
    this state provides:
        is_in_pregame()
        next_player()
        begin_turn()
        has_rolled()
        can_roll()
        can_move_robber()
        can_steal()
        can_buy_road()
        can_buy_settlement()
        can_buy_city()
        can_buy_dev_card()
        can_trade()
        can_play_knight()
        can_play_monopoly()
        can_play_road_builder()
        can_play_victory_point()
    sub-states must implement:
        can_end_turn()
    """
    def is_in_game(self):
        return True

    def is_in_pregame(self):
        """
        See GameStatePreGame for details.

        :return: Boolean
        """
        return False

    def next_player(self):
        """
        Returns the player whose turn it will be next.

        Uses regular seat-wise clockwise rotation.

        Compare to GameStatePreGame's implementation, which uses snake draft.

        :return Player
        """
        logging.warning('turn={}, players={}'.format(
            self.game._cur_turn,
            self.game.players
        ))
        return self.game.players[(self.game._cur_turn + 1) % len(self.game.players)]

    def begin_turn(self):
        """
        Begins the turn for the current player.

        All that is required is to set the game's state.

        Compare to GameStatePreGame's implementation, which uses GameStatePreGamePlaceSettlement

        :return None
        """
        self.game.set_state(GameStateBeginTurn(self.game))

    def has_rolled(self):
        """
        Whether the current player has rolled or not.

        :return Boolean
        """
        return self.game.last_player_to_roll == self.game.get_cur_player()

    def can_roll(self):
        """
        Whether the current player can roll or not.

        A player can roll only if they have not yet rolled.

        :return Boolean
        """
        return not self.has_rolled()

    def can_move_robber(self):
        """
        Whether the current player can move the robber or not.

        :return Boolean
        """
        return False

    def can_steal(self):
        """
        Whether the current player can steal or not.

        :return Boolean
        """
        return False

    def can_buy_road(self):
        """
        Whether the current player can buy a road or not.

        :return Boolean
        """
        return self.has_rolled()

    def can_buy_settlement(self):
        """
        Whether the current player can buy a settlement or not.

        :return Boolean
        """
        return self.has_rolled()

    def can_buy_city(self):
        """
        Whether the current player can buy a city or not.

        :return Boolean
        """
        return self.has_rolled()

    def can_place_road(self):
        """
        Whether the current player can place a road or not.

        :return Boolean
        """
        return False

    def can_place_settlement(self):
        """
        Whether the current player can place a settlement or not.

        :return Boolean
        """
        return False

    def can_place_city(self):
        """
        Whether the current player can place a city or not.

        :return Boolean
        """
        return False

    def can_buy_dev_card(self):
        """
        Whether the current player can buy a dev card or not.

        :return Boolean
        """
        return self.has_rolled()

    def can_trade(self):
        """
        Whether the current player can trade or not.

        :return Boolean
        """
        return self.has_rolled()

    def can_play_knight(self):
        """
        Whether the current player can play a knight dev card or not.

        :return Boolean
        """
        return self.game.dev_card_state.can_play_dev_card()

    def can_play_monopoly(self):
        """
        Whether the current player can play a monopoly dev card or not.

        :return Boolean
        """
        return self.has_rolled() and self.game.dev_card_state.can_play_dev_card()

    def can_play_year_of_plenty(self):
        """
        Whether the current player can play a year of plenty dev card or not.

        :return Boolean
        """
        return self.has_rolled() and self.game.dev_card_state.can_play_dev_card()

    def can_play_road_builder(self):
        """
        Whether the current player can play a road builder dev card or not.

        :return Boolean
        """
        return self.has_rolled() and self.game.dev_card_state.can_play_dev_card()

    def can_play_victory_point(self):
        """
        Whether the current player can play a victory point dev card or not.

        :return Boolean
        """
        return True

    def can_end_turn(self):
        """
        Whether the current player can end their turn or not.

        :return Boolean
        """
        raise NotImplemented()


class GameStatePreGame(GameStateInGame):
    """
    The pregame is defined as
    - AFTER the board has been laid out
    - BEFORE the first dice roll

    In other words, it is the placing of the initial settlements and roads, in snake draft order.

    this state implements:
        can_end_turn()

    this state provides:
        None
    sub-classes must implement:
        None
    """
    def can_end_turn(self):
        return False

    def is_in_pregame(self):
        return True

    def next_player(self):
        snake = self.game.players.copy()
        snake += list(reversed(snake))
        try:
            return snake[self.game._cur_turn + 1]
        except IndexError:
            self.game.set_state(GameStateBeginTurn(self.game))
            return self.game.state.next_player()

    def begin_turn(self):
        self.game.set_state(GameStatePreGamePlaceSettlement(self.game))

    def can_play_knight(self):
        """No dev cards in the pregame"""
        return False

    def can_play_monopoly(self):
        """No dev cards in the pregame"""
        return False

    def can_play_road_builder(self):
        """No dev cards in the pregame"""
        return False

    def can_play_victory_point(self):
        """No dev cards in the pregame"""
        return False

    def can_roll(self):
        """No rolling in the pregame"""
        return False

    def can_buy_road(self):
        raise NotImplemented()

    def can_buy_settlement(self):
        raise NotImplemented()

    def can_buy_city(self):
        """No cities in the pregame"""
        return False

    def can_buy_dev_card(self):
        """No dev cards in the pregame"""
        return False

    def can_trade(self):
        """No trading in the pregame"""
        return False


class GameStatePreGamePlaceSettlement(GameStatePreGame):
    """
    - AFTER a player's turn has started
    - BEFORE the player has placed an initial settlement
    """
    def can_buy_settlement(self):
        return True

    def can_buy_road(self):
        return False

    def can_end_turn(self):
        return False


class GameStatePreGamePlaceRoad(GameStatePreGame):
    """
    - AFTER a player has placed an initial settlement
    - BEFORE the player has placed an initial road
    """
    def can_buy_settlement(self):
        return False

    def can_buy_road(self):
        return True

    def can_end_turn(self):
        return False


class GameStatePreGamePlacingPiece(GameStatePreGame):
    """
    - AFTER a player has selected to place a piece
    - WHILE the player is choosing where to place it
    - BEFORE the player has placed it
    """
    def __init__(self, game, piece_type):
        super(GameStatePreGamePlacingPiece, self).__init__(game)
        self.piece_type = piece_type

    def can_buy_settlement(self):
        return False

    def can_buy_road(self):
        return False

    def can_end_turn(self):
        return False

    def can_place_road(self):
        return self.piece_type == catan.pieces.PieceType.road

    def can_place_settlement(self):
        return self.piece_type == catan.pieces.PieceType.settlement

    def can_place_city(self):
        return self.piece_type == catan.pieces.PieceType.city

    def place_road(self, edge):
        if not self.can_place_road():
            logging.warning('Attempted to place road in illegal state={} with piece_type={}'.format(
                self.__class__.__name__,
                self.piece_type
            ))
        self.game.buy_road(edge)

    def place_settlement(self, node):
        if not self.can_place_settlement():
            logging.warning('Attempted to place settlement in illegal state={} with piece_type={}'.format(
                self.__class__.__name__,
                self.piece_type
            ))
        self.game.buy_settlement(node)

    def place_city(self, node):
        if not self.can_place_city():
            logging.warning('Attempted to place city in illegal state={} with piece_type={}'.format(
                self.__class__.__name__,
                self.piece_type
            ))
        self.game.buy_city(node)

class GameStateBeginTurn(GameStateInGame):
    """
    The start of the turn is defined as
    - AFTER the previous player ends their turn
    - BEFORE the next player's first action
    """
    def can_end_turn(self):
        return False


class GameStateMoveRobber(GameStateInGame):
    """
    Defined as
    - AFTER the rolling of a 7
    - BEFORE the player has moved the robber
    """
    def can_end_turn(self):
        return False

    def can_move_robber(self):
        return True

    def move_robber(self, tile_id):
        robbers = self.game.board.get_pieces((catan.pieces.PieceType.robber, ),
                                             hexgrid.tile_id_to_coord(self.game.robber_tile))
        for robber in robbers:
            self.game.board.move_piece(robber,
                                       hexgrid.tile_id_to_coord(self.game.robber_tile), hexgrid.tile_id_to_coord(tile_id))
        if len(robbers) != 1:
            logging.warning('{} robbers found in board.pieces'.format(len(robbers)))
        self.game.robber_tile = tile_id
        self.game.set_state(GameStateSteal(self.game))

    def can_roll(self):
        return False

    def can_buy_road(self):
        return False

    def can_buy_settlement(self):
        return False

    def can_buy_city(self):
        return False

    def can_buy_dev_card(self):
        return False

    def can_trade(self):
        return False

    def can_play_knight(self):
        return False

    def can_play_monopoly(self):
        return False

    def can_play_road_builder(self):
        return False


class GameStateMoveRobberUsingKnight(GameStateMoveRobber):
    """
    Defined as
    - AFTER the playing of a knight
    - BEFORE the player has moved the robber
    """
    def move_robber(self, tile_id):
        robbers = self.game.board.get_pieces((catan.pieces.PieceType.robber, ),
                                             hexgrid.tile_id_to_coord(self.game.robber_tile))
        for robber in robbers:
            self.game.board.move_piece(robber,
                                       hexgrid.tile_id_to_coord(self.game.robber_tile), hexgrid.tile_id_to_coord(tile_id))
        if len(robbers) > 1:
            logging.warning('More than one robber found in board.pieces')
        self.game.robber_tile = tile_id
        self.game.set_state(GameStateStealUsingKnight(self.game))


class GameStateSteal(GameStateInGame):
    """
    Defined as
    - AFTER the player has moved the robber
    - BEFORE the player has stolen a card
    """
    def can_end_turn(self):
        return False

    def can_steal(self):
        return True

    def steal(self, victim):
        self.game.catanlog.log_player_moves_robber_and_steals(
            self.game.get_cur_player(),
            self.game.robber_tile,
            victim
        )
        self.game.set_state(GameStateDuringTurnAfterRoll(self.game))

    def can_roll(self):
        return False

    def can_buy_road(self):
        return False

    def can_buy_settlement(self):
        return False

    def can_buy_city(self):
        return False

    def can_buy_dev_card(self):
        return False

    def can_trade(self):
        return False

    def can_play_knight(self):
        return False

    def can_play_monopoly(self):
        return False

    def can_play_road_builder(self):
        return False


class GameStateStealUsingKnight(GameStateSteal):
    """
    Defined as
    - AFTER the player has moved the robber using the knight
    - BEFORE the player has stolen a card using the knight
    """
    def steal(self, victim):
        self.game.catanlog.log_player_plays_dev_knight(
            self.game.get_cur_player(),
            self.game.robber_tile,
            victim
        )
        self.game.set_state(GameStateDuringTurnAfterRoll(self.game))


class GameStateDuringTurnAfterRoll(GameStateInGame):
    """
    The most common state.

    Defined as
    - AFTER the player's roll
    - BEFORE the player ends their turn
    """
    def can_end_turn(self):
        return True


class GameStatePlacingPiece(GameStateInGame):
    """
    - AFTER a player has selected to place a piece
    - WHILE the player is choosing where to place it
    - BEFORE the player has placed it
    """
    def __init__(self, game, piece_type):
        super(GameStatePlacingPiece, self).__init__(game)
        self.piece_type = piece_type

    def can_end_turn(self):
        return False

    def can_place_road(self):
        return self.piece_type == catan.pieces.PieceType.road

    def can_place_settlement(self):
        return self.piece_type == catan.pieces.PieceType.settlement

    def can_place_city(self):
        return self.piece_type == catan.pieces.PieceType.city

    def place_road(self, edge):
        if not self.can_place_road():
            logging.warning('Attempted to place road in illegal state={} with piece_type={}'.format(
                self.__class__.__name__,
                self.piece_type
            ))
        self.game.buy_road(edge)

    def place_settlement(self, node):
        if not self.can_place_settlement():
            logging.warning('Attempted to place settlement in illegal state={} with piece_type={}'.format(
                self.__class__.__name__,
                self.piece_type
            ))
        self.game.buy_settlement(node)

    def place_city(self, node):
        if not self.can_place_city():
            logging.warning('Attempted to place city in illegal state={} with piece_type={}'.format(
                self.__class__.__name__,
                self.piece_type
            ))
        self.game.buy_city(node)

    ###

    def can_move_robber(self):
        return False

    def can_steal(self):
        return False

    def can_buy_road(self):
        return False

    def can_buy_settlement(self):
        return False

    def can_buy_city(self):
        return False

    def can_buy_dev_card(self):
        return False

    def can_trade(self):
        return False

    def can_play_knight(self):
        return False

    def can_play_monopoly(self):
        return False

    def can_play_road_builder(self):
        return False

    def can_play_victory_point(self):
        return True


class GameStatePlacingRoadBuilderPieces(GameStatePlacingPiece):
    """
    - AFTER a player has selected to build 2 road builder roads
    - WHILE the player is choosing where to place them
    - BEFORE the player has placed both of them
    """
    def __init__(self, game):
        super(GameStatePlacingRoadBuilderPieces, self).__init__(game, catan.pieces.PieceType.road)
        self.edges = list()

    def place_road(self, edge):
        if not self.can_place_road():
            logging.warning('Attempted to place road in illegal state={} with piece_type={}'.format(
                self.__class__.__name__,
                self.piece_type
            ))
        piece = catan.pieces.Piece(catan.pieces.PieceType.road, self.game.get_cur_player())
        self.game.board.place_piece(piece, edge)
        self.edges.append(edge)
        if len(self.edges) == 2:
            self.game.play_road_builder(self.edges[0], self.edges[1])
            self.game.set_state(GameStateDuringTurnAfterRoll(self.game))


class DevCardPlayabilityState(object):
    def __init__(self, game):
        self.game = game

    def can_play_dev_card(self):
        raise NotImplemented()


class DevCardNotPlayedState(DevCardPlayabilityState):
    def can_play_dev_card(self):
        return True


class DevCardPlayedState(DevCardPlayabilityState):
    def can_play_dev_card(self):
        return False


class BoardState(object):
    def __init__(self, board):
        self.board = board

    def modifiable(self):
        raise NotImplemented()


class BoardStateModifiable(BoardState):
    def modifiable(self):
        return True


class BoardStateLocked(BoardState):
    def modifiable(self):
        return False



import copy
import logging

import hexgrid
import catanlog
import undoredo

import catan.states
import catan.board
import catan.pieces


class Game(object):
    """
    class Game represents a single game of catan. It has players, a board, and a log.

    A Game has observers. Observers register themselves by adding themselves to
    the Game's observers set. When the Game changes, it will notify all its observers,
    who can then poll the game state and make changes accordingly.

    e.g. self.game.observers.add(self)

    A Game has state. When changing state, remember to pass the current game to the
    state's constructor. This allows the state to modify the game as appropriate in
    the current state.

    e.g. self.set_state(states.GameStateNotInGame(self))
    """
    def __init__(self, players=None, board=None, logging='on', pregame='on'):
        """
        Create a Game with the given options.

        :param players: list(Player)
        :param board: Board
        :param logging: (on|off)
        :param pregame: (on|off)
        """
        self.observers = set()
        self.undo_manager = undoredo.UndoManager()
        self.options = {
            'pregame': pregame,
        }
        self.players = players or list()
        self.board = board or catan.board.Board()
        self.robber = catan.pieces.Piece(catan.pieces.PieceType.robber, None)
        if logging == 'on':
            self.catanlog = catanlog.CatanLog()
        else:
            self.catanlog = catanlog.NoopCatanLog()

        self.state = None # set in #set_state
        self.dev_card_state = None # set in #set_dev_card_state
        self._cur_player = None # set in #set_players
        self.last_roll = None # set in #roll
        self.last_player_to_roll = None # set in #roll
        self._cur_turn = 0 # incremented in #end_turn
        self.robber_tile = None # set in #move_robber

        self.board.observers.add(self)

        self.set_state(catan.states.GameStateNotInGame(self))
        self.set_dev_card_state(catan.states.DevCardNotPlayedState(self))

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == 'observers':
                setattr(result, k, set(v))
            elif k == 'state':
                setattr(result, k, v)
            elif k == 'undo_manager':
                setattr(result, k, v)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def do(self, command: undoredo.Command):
        """
        Does the command using the undo_manager's stack
        :param command: Command
        """
        self.undo_manager.do(command)
        self.notify_observers()

    def undo(self):
        """
        Rewind the game to the previous state.
        """
        self.undo_manager.undo()
        self.notify_observers()
        logging.debug('undo_manager undo stack={}'.format(self.undo_manager._undo_stack))

    def redo(self):
        """
        Redo the latest undone command.
        """
        self.undo_manager.redo()
        self.notify_observers()
        logging.debug('undo_manager redo stack={}'.format(self.undo_manager._redo_stack))

    def copy(self):
        """
        Return a deep copy of this Game object. See Game.__deepcopy__ for the copy implementation.
        :return: Game
        """
        return copy.deepcopy(self)

    def restore(self, game):
        """
        Restore this Game object to match the properties and state of the given Game object
        :param game: properties to restore to the current (self) Game
        """
        self.observers = game.observers
        # self.undo_manager = game.undo_manager
        self.options = game.options
        self.players = game.players
        self.board.restore(game.board)
        self.robber = game.robber
        self.catanlog = game.catanlog

        self.state = game.state
        self.state.game = self

        self.dev_card_state = game.dev_card_state

        self._cur_player = game._cur_player
        self.last_roll = game.last_roll
        self.last_player_to_roll = game.last_player_to_roll
        self._cur_turn = game._cur_turn
        self.robber_tile = game.robber_tile

        self.notify_observers()

    def notify(self, observable):
        self.notify_observers()

    def notify_observers(self):
        for obs in self.observers.copy():
            obs.notify(self)

    def set_state(self, game_state):
        _old_state = self.state
        _old_board_state = self.board.state
        self.state = game_state
        if game_state.is_in_game():
            self.board.lock()
        else:
            self.board.unlock()
        logging.info('Game now={}, was={}. Board now={}, was={}'.format(
            type(self.state).__name__,
            type(_old_state).__name__,
            type(self.board.state).__name__,
            type(_old_board_state).__name__
        ))
        self.notify_observers()

    def set_dev_card_state(self, dev_state):
        self.dev_card_state = dev_state
        self.notify_observers()

    @undoredo.undoable
    def start(self, players):
        """
        Start the game.

        The value of option 'pregame' determines whether the pregame will occur or not.

        - Resets the board
        - Sets the players
        - Sets the game state to the appropriate first turn of the game
        - Finds the robber on the board, sets the robber_tile appropriately
        - Logs the catanlog header

        :param players: players to start the game with
        """
        from .boardbuilder import Opt
        self.reset()
        if self.board.opts.get('players') == Opt.debug:
            players = Game.get_debug_players()
        self.set_players(players)
        if self.options.get('pregame') is None or self.options.get('pregame') == 'on':
            logging.debug('Entering pregame, game options={}'.format(self.options))
            self.set_state(catan.states.GameStatePreGamePlaceSettlement(self))
        elif self.options.get('pregame') == 'off':
            logging.debug('Skipping pregame, game options={}'.format(self.options))
            self.set_state(catan.states.GameStateBeginTurn(self))

        terrain = list()
        numbers = list()
        for tile in self.board.tiles:
            terrain.append(tile.terrain)
            numbers.append(tile.number)

        for (_, coord), piece in self.board.pieces.items():
            if piece.type == catan.pieces.PieceType.robber:
                self.robber_tile = hexgrid.tile_id_from_coord(coord)
                logging.debug('Found robber at coord={}, set robber_tile={}'.format(coord, self.robber_tile))

        self.catanlog.log_game_start(self.players, terrain, numbers, self.board.ports)
        self.notify_observers()

    def end(self):
        self.catanlog.log_player_wins(self.get_cur_player())
        self.set_state(catan.states.GameStateNotInGame(self))

    def reset(self):
        self.players = list()
        self.state = catan.states.GameStateNotInGame(self)

        self.last_roll = None
        self.last_player_to_roll = None
        self._cur_player = None
        self._cur_turn = 0

        self.notify_observers()

    def get_cur_player(self):
        if self._cur_player is None:
            return Player(1, 'nobody', 'nobody')
        else:
            return Player(self._cur_player.seat, self._cur_player.name, self._cur_player.color)

    def set_cur_player(self, player):
        self._cur_player = Player(player.seat, player.name, player.color)

    def set_players(self, players):
        self.players = list(players)
        self.set_cur_player(self.players[0])
        self.notify_observers()

    def cur_player_has_port_type(self, port_type):
        return self.player_has_port_type(self.get_cur_player(), port_type)

    def player_has_port_type(self, player, port_type):
        for port in self.board.ports:
            if port.type == port_type and self._player_has_port(player, port):
                return True
        return False

    def _player_has_port(self, player, port):
        edge_coord = hexgrid.edge_coord_in_direction(port.tile_id, port.direction)
        for node in hexgrid.nodes_touching_edge(edge_coord):
            pieces = self.board.get_pieces((catan.pieces.PieceType.settlement, catan.pieces.PieceType.city), node)
            if len(pieces) < 1:
                continue
            elif len(pieces) > 1:
                raise Exception('Probably a bug, num={} pieces found on node={}'.format(
                    len(pieces), node
                ))
            assert len(pieces) == 1  # will be asserted by previous if/elif combo
            piece = pieces[0]
            if piece.owner == player:
                return True
        return False

    @undoredo.undoable
    def roll(self, roll):
        self.catanlog.log_player_roll(self.get_cur_player(), roll)
        self.last_roll = roll
        self.last_player_to_roll = self.get_cur_player()
        if int(roll) == 7:
            self.set_state(catan.states.GameStateMoveRobber(self))
        else:
            self.set_state(catan.states.GameStateDuringTurnAfterRoll(self))

    @undoredo.undoable
    def move_robber(self, tile):
        self.state.move_robber(tile)

    @undoredo.undoable
    def steal(self, victim):
        if victim is None:
            victim = Player(1, 'nobody', 'nobody')
        self.state.steal(victim)

    def stealable_players(self):
        if self.robber_tile is None:
            return list()
        stealable = set()
        for node in hexgrid.nodes_touching_tile(self.robber_tile):
            pieces = self.board.get_pieces(types=(catan.pieces.PieceType.settlement, catan.pieces.PieceType.city), coord=node)
            if pieces:
                logging.debug('found stealable player={}, cur={}'.format(pieces[0].owner, self.get_cur_player()))
                stealable.add(pieces[0].owner)
        if self.get_cur_player() in stealable:
            stealable.remove(self.get_cur_player())
        logging.debug('stealable players={} at robber tile={}'.format(stealable, self.robber_tile))
        return stealable

    @undoredo.undoable
    def begin_placing(self, piece_type):
        if self.state.is_in_pregame():
            self.set_state(catan.states.GameStatePreGamePlacingPiece(self, piece_type))
        else:
            self.set_state(catan.states.GameStatePlacingPiece(self, piece_type))

    # @undoredo.undoable # state.place_road calls this, place_road is undoable
    def buy_road(self, edge):
        #self.assert_legal_road(edge)
        piece = catan.pieces.Piece(catan.pieces.PieceType.road, self.get_cur_player())
        self.board.place_piece(piece, edge)
        self.catanlog.log_player_buys_road(self.get_cur_player(), edge)
        if self.state.is_in_pregame():
            self.end_turn()
        else:
            self.set_state(catan.states.GameStateDuringTurnAfterRoll(self))

    # @undoredo.undoable # state.place_settlement calls this, place_settlement is undoable
    def buy_settlement(self, node):
        #self.assert_legal_settlement(node)
        piece = catan.pieces.Piece(catan.pieces.PieceType.settlement, self.get_cur_player())
        self.board.place_piece(piece, node)
        self.catanlog.log_player_buys_settlement(self.get_cur_player(), node)
        if self.state.is_in_pregame():
            self.set_state(catan.states.GameStatePreGamePlaceRoad(self))
        else:
            self.set_state(catan.states.GameStateDuringTurnAfterRoll(self))

    # @undoredo.undoable # state.place_city calls this, place_city is undoable
    def buy_city(self, node):
        #self.assert_legal_city(node)
        piece = catan.pieces.Piece(catan.pieces.PieceType.city, self.get_cur_player())
        self.board.place_piece(piece, node)
        self.catanlog.log_player_buys_city(self.get_cur_player(), node)
        self.set_state(catan.states.GameStateDuringTurnAfterRoll(self))

    @undoredo.undoable
    def buy_dev_card(self):
        self.catanlog.log_player_buys_dev_card(self.get_cur_player())
        self.notify_observers()

    @undoredo.undoable
    def place_road(self, edge_coord):
        self.state.place_road(edge_coord)

    @undoredo.undoable
    def place_settlement(self, node_coord):
        self.state.place_settlement(node_coord)

    @undoredo.undoable
    def place_city(self, node_coord):
        self.state.place_city(node_coord)

    @undoredo.undoable
    def trade(self, trade):
        giver = trade.giver().color
        giving = [(n, t.value) for n, t in trade.giving()]
        getting = [(n, t.value) for n, t in trade.getting()]
        if trade.getter() in catan.board.PortType:
            getter = trade.getter().value
            self.catanlog.log_player_trades_with_port(giver, giving, getter, getting)
            logging.debug('trading {} to port={} to get={}'.format(giving, getter, getting))
        else:
            getter = trade.getter().color
            self.catanlog.log_player_trades_with_other(giver, giving, getter, getting)
            logging.debug('trading {} to player={} to get={}'.format(giving, getter, getting))
        self.notify_observers()

    @undoredo.undoable
    def play_knight(self):
        self.set_dev_card_state(catan.states.DevCardPlayedState(self))
        self.set_state(catan.states.GameStateMoveRobberUsingKnight(self))

    @undoredo.undoable
    def play_monopoly(self, resource):
        self.catanlog.log_player_plays_dev_monopoly(self.get_cur_player(), resource)
        self.set_dev_card_state(catan.states.DevCardPlayedState(self))

    @undoredo.undoable
    def play_year_of_plenty(self, resource1, resource2):
        self.catanlog.log_player_plays_dev_year_of_plenty(self.get_cur_player(), resource1, resource2)
        self.set_dev_card_state(catan.states.DevCardPlayedState(self))

    @undoredo.undoable
    def play_road_builder(self, edge1, edge2):
        self.catanlog.log_player_plays_dev_road_builder(self.get_cur_player(), edge1, edge2)
        self.set_dev_card_state(catan.states.DevCardPlayedState(self))

    @undoredo.undoable
    def play_victory_point(self):
        self.catanlog.log_player_plays_dev_victory_point(self.get_cur_player())
        self.set_dev_card_state(catan.states.DevCardPlayedState(self))

    @undoredo.undoable
    def end_turn(self):
        self.catanlog.log_player_ends_turn(self.get_cur_player())
        self.set_cur_player(self.state.next_player())
        self._cur_turn += 1

        self.set_dev_card_state(catan.states.DevCardNotPlayedState(self))
        if self.state.is_in_pregame():
            self.set_state(catan.states.GameStatePreGamePlaceSettlement(self))
        else:
            self.set_state(catan.states.GameStateBeginTurn(self))

    @classmethod
    def get_debug_players(cls):
        return [Player(1, 'yurick', 'green'),
                Player(2, 'josh', 'blue'),
                Player(3, 'zach', 'orange'),
                Player(4, 'ross', 'red')]


class Player(object):
    """class Player represents a single player on the game board.

    :param seat: integer, with 1 being top left, and increasing clockwise
    :param name: will be lowercased, spaces will be removed
    :param color: will be lowercased, spaces will be removed
    """
    def __init__(self, seat, name, color):
        if not (1 <= seat <= 4):
            raise Exception("Seat must be on [1,4]")
        self.seat = seat

        self.name = name.lower().replace(' ', '')
        self.color = color.lower().replace(' ', '')

    def __eq__(self, other):
        if other is None:
            return False
        if other.__class__ != Player:
            return False
        return (self.color == other.color
                and self.name == other.name
                and self.seat == other.seat)

    def __repr__(self):
        return '{} ({})'.format(self.color, self.name)

    def __hash__(self):
        return sum(bytes(str(self), encoding='utf8'))




import logging
from collections import Counter


class CatanTrade(object):
    """
    class CatanTrade provides a mutable trade object for catan

    The trade relationship is one-to-one, and supports any number of
    each of the resources going in both directions.

    Usually, the current player is the giver, and the other entity the getter.
    Think of it as: the current player gives resources, and gets some in return.

    Use give() and get() to add resources to the trade.

    Resources cannot be removed from the trade. If you want this functionality,
    delete the trade and build a new one instead.
    """
    def __init__(self, giver=None, getter=None):
        self._give = list()
        self._get = list()
        self._giver = giver
        self._getter = getter

    def give(self, terrain, num=1):
        """
        Add a certain number of resources to the trade from giver->getter
        :param terrain: resource type, models.Terrain
        :param num: number to add, int
        :return: None
        """
        for _ in range(num):
            logging.debug('terrain={}'.format(terrain))
            self._give.append(terrain)

    def get(self, terrain, num=1):
        """
        Add a certain number of resources to the trade from getter->giver
        :param terrain: resource type, models.Terrain
        :param num: number to add, int
        :return: None
        """
        for _ in range(num):
            logging.debug('terrain={}'.format(terrain))
            self._get.append(terrain)

    def giver(self):
        return self._giver

    def getter(self):
        return self._getter

    def giving(self):
        """
        Returns tuples corresponding to the number and type of each
        resource in the trade from giver->getter

        :return: eg [(2, Terrain.wood), (1, Terrain.brick)]
        """
        logging.debug('give={}'.format(self._give))
        c = Counter(self._give.copy())
        return [(n, t) for t, n in c.items()]

    def getting(self):
        """
        Returns tuples corresponding to the number and type of each
        resource in the trade from getter->giver

        :return: eg [(2, Terrain.wood), (1, Terrain.brick)]
        """
        c = Counter(self._get.copy())
        return [(n, t) for t, n in c.items()]

    def num_giving(self):
        return len(self._give)

    def num_getting(self):
        return len(self._get)

    def set_giver(self, giver):
        self._giver = giver

    def set_getter(self, getter):
        self._getter = getter



from distutils.core import setup

with open("README.md", "r") as fp:
    long_description = fp.read()

version = dict()
with open('version.py', 'r') as fp:
    exec(fp.read(), version)

setup(name="catan",
      version=version['__version__'],
      author="Ross Anderson",
      author_email="ross.anderson@ualberta.ca",
      url="https://github.com/rosshamish/catan-py/",
      download_url = 'https://github.com/rosshamish/catan-py/tarball/' + version['__version__'],
      description="models for representing and manipulating a game of catan",
      long_description=long_description,
      keywords=[],
      classifiers=[],
      license="GPLv3",

      packages=["catan"],
      install_requires=[
          'hexgrid',
          'catanlog',
          'undoredo',
      ],
	)

